from argparse import ArgumentParser, Action, SUPPRESS, \
    HelpFormatter, RawDescriptionHelpFormatter, ArgumentTypeError
from pathlib import Path
import sys
import os

from portablemc.standard import Context, VersionManifest
from portablemc.auth import AuthDatabase

from .util import LibrarySpecifierFilter
from .output import Output
from .lang import get as _

from typing import Optional, Type, Tuple, List, Dict

# The following classes are only used for type checking and represent a typed namespace
# as produced by the arguments registered to the argument parser.

class RootNs:
    main_dir: Optional[Path]
    work_dir: Optional[Path]
    timeout: float
    out_kind: str
    verbose: int
    # Initialized by main function after argument parsing.
    parser: ArgumentParser
    out: Output
    context: Context
    version_manifest: VersionManifest
    auth_database: AuthDatabase
    socket_error_tips: List[str]

class SearchNs(RootNs):
    kind: str
    input: str

class AuthBaseNs(RootNs):
    auth_service: str
    auth_no_browser: bool

class StartNs(AuthBaseNs):
    dry: bool
    disable_mp: bool
    disable_chat: bool
    demo: bool
    resolution: Optional[Tuple[int, int]]
    jvm: Optional[Path]
    jvm_args: Optional[str]
    no_fix: bool
    babric_prefix: str
    fabric_prefix: str
    legacyfabric_prefix: str
    quilt_prefix: str
    forge_prefix: str
    neoforge_prefix: str
    lwjgl: Optional[str]
    exclude_lib: Optional[List[LibrarySpecifierFilter]]
    include_bin: Optional[List[Path]]
    temp_login: bool
    auth_service: str
    auth_anonymize: bool
    login: Optional[str]
    username: Optional[str]
    uuid: Optional[str]
    server: Optional[str]
    server_port: Optional[int]
    version: str

class LoginNs(AuthBaseNs):
    email_or_username: str

class LogoutNs(AuthBaseNs):
    email_or_username: str

class ShowCompletionNs(RootNs):
    shell: str


def register_common_help(parser: ArgumentParser) -> None:
    parser.formatter_class = new_help_formatter_class(40)
    parser.add_argument("-h", "--help", action="help", default=SUPPRESS, help=_("args.common.help"))


def register_common_auth_service(parser: ArgumentParser) -> None:
    
    auth_choices = get_auth_services()
    auth_arg = parser.add_argument("--auth-service", help=_("args.common.auth_service"), default="microsoft", choices=auth_choices)
    for choice in auth_choices:
        add_completion(auth_arg, choice, _(f"args.common.auth_service.comp.{choice}"))

    parser.add_argument("--auth-no-browser", help=_("args.common.auth_no_browser"), action="store_true")


def register_arguments() -> ArgumentParser:

    parser = ArgumentParser(allow_abbrev=False, prog="portablemc", description=_("args._"), add_help=False)
    register_common_help(parser)
    
    parser.add_argument("--main-dir", help=_("args.main_dir"), type=type_path_dir)
    parser.add_argument("--work-dir", help=_("args.work_dir"), type=type_path_dir)
    parser.add_argument("--timeout", help=_("args.timeout"), type=float)

    output_choices = get_outputs()
    output_default = "human-color" if sys.stdout.isatty() else "human"
    output_arg = parser.add_argument("--output", help=_("args.output"), dest="out_kind", choices=output_choices, default=output_default)
    for choice in output_choices:
        add_completion(output_arg, choice, _(f"args.output.comp.{choice}"))

    parser.add_argument("-v", dest="verbose", help=_("args.verbose"), action="count", default=0)
    register_subcommands(parser.add_subparsers(title="subcommands", dest="subcommand"))

    return parser


def register_subcommands(subparsers) -> None:
    register_search_arguments(subparsers.add_parser("search", help=_("args.search"), add_help=False))
    register_start_arguments(subparsers.add_parser("start", help=_("args.start"), add_help=False))
    register_login_arguments(subparsers.add_parser("login", help=_("args.login"), add_help=False))
    register_logout_arguments(subparsers.add_parser("logout", help=_("args.logout"), add_help=False))
    register_show_arguments(subparsers.add_parser("show", help=_("args.show"), add_help=False))


def register_search_arguments(parser: ArgumentParser) -> None:
    
    parser.description = _("args.search._")
    register_common_help(parser)

    kind_choices = get_search_kinds()
    kind_arg = parser.add_argument("-k", "--kind", help=_("args.search.kind"), default="mojang", choices=kind_choices)
    for choice in kind_choices:
        add_completion(kind_arg, choice, _(f"args.search.kind.comp.{choice}"))

    input_arg = parser.add_argument("input", nargs="?", help=_("args.search.input"))
    add_completion(input_arg, "release", _("args.search.input.comp.release"))
    add_completion(input_arg, "snapshot", _("args.search.input.comp.snapshot"))

def register_start_arguments(parser: ArgumentParser) -> None:
    
    register_common_help(parser)
    parser.add_argument("--dry", help=_("args.start.dry"), action="store_true")
    parser.add_argument("--disable-mp", help=_("args.start.disable_multiplayer"), action="store_true")
    parser.add_argument("--disable-chat", help=_("args.start.disable_chat"), action="store_true")
    parser.add_argument("--demo", help=_("args.start.demo"), action="store_true")
    parser.add_argument("--resolution", help=_("args.start.resolution"), type=type_resolution)
    parser.add_argument("--jvm", help=_("args.start.jvm"), type=type_path)
    parser.add_argument("--jvm-args", help=_("args.start.jvm_args"), metavar="ARGS")
    parser.add_argument("--no-fix", help=_("args.start.no_fix"), action="store_true")
    parser.add_argument("--fabric-prefix", help=_("args.start.fabric_prefix"), default="fabric", metavar="PREFIX")
    parser.add_argument("--quilt-prefix", help=_("args.start.quilt_prefix"), default="quilt", metavar="PREFIX")
    parser.add_argument("--legacyfabric-prefix", help=_("args.start.legacyfabric_prefix"), default="legacyfabric", metavar="PREFIX")
    parser.add_argument("--babric-prefix", help=_("args.start.babric_prefix"), default="babric", metavar="PREFIX")
    parser.add_argument("--forge-prefix", help=_("args.start.forge_prefix"), default="forge", metavar="PREFIX")
    parser.add_argument("--neoforge-prefix", help=_("args.start.neoforge_prefix"), default="neoforge", metavar="PREFIX")
    parser.add_argument("--lwjgl", help=_("args.start.lwjgl"))
    parser.add_argument("--exclude-lib", help=_("args.start.exclude_lib"), action="append", metavar="SPEC", type=LibrarySpecifierFilter.from_str)
    parser.add_argument("--include-bin", help=_("args.start.include_bin"), action="append", metavar="PATH", type=type_path)
    parser.add_argument("--auth-anonymize", help=_("args.start.auth_anonymize"), action="store_true")
    register_common_auth_service(parser)
    parser.add_argument("-t", "--temp-login", help=_("args.start.temp_login"), action="store_true")
    parser.add_argument("-l", "--login", help=_("args.start.login"), type=type_email_or_username)
    parser.add_argument("-u", "--username", help=_("args.start.username"), metavar="NAME")
    parser.add_argument("-i", "--uuid", help=_("args.start.uuid"))
    parser.add_argument("-s", "--server", help=_("args.start.server"), type=type_host)
    parser.add_argument("-p", "--server-port", help=_("args.start.server_port"), metavar="PORT")

    version_arg = parser.add_argument("version", nargs="?", default="release", help=_("args.start.version", formats=", ".join(map(lambda s: _(f"args.start.version.{s}"), ("standard", "fabric", "quilt", "legacyfabric", "babric", "forge", "neoforge")))))
    for standard in ("release", "snapshot"):
        add_completion(version_arg, standard, _(f"args.start.version.comp.{standard}"))
    for loader in ("fabric", "quilt", "legacyfabric", "babric", "forge", "neoforge"):
        add_completion(version_arg, f"{loader}:", _(f"args.start.version.comp.{loader}"))


def register_login_arguments(parser: ArgumentParser) -> None:
    register_common_help(parser)
    register_common_auth_service(parser)
    parser.add_argument("email_or_username", type=type_email_or_username)


def register_logout_arguments(parser: ArgumentParser) -> None:
    register_common_help(parser)
    register_common_auth_service(parser)
    parser.add_argument("email_or_username", type=type_email_or_username)


def register_show_arguments(parser: ArgumentParser) -> None:
    register_common_help(parser)
    subparsers = parser.add_subparsers(title="subcommands", dest="show_subcommand")
    subparsers.required = True
    subparsers.add_parser("about", help=_("args.show.about"), add_help=False)
    subparsers.add_parser("auth", help=_("args.show.auth"), add_help=False)
    subparsers.add_parser("lang", help=_("args.show.lang"), add_help=False)
    register_show_completion_arguments(subparsers.add_parser("completion", help=_("args.show.completion"), add_help=False))


def register_show_completion_arguments(parser: ArgumentParser) -> None:

    parser.description = _("args.show.completion._")
    register_common_help(parser)

    # The shell argument is only required if the shell cannot be determined.
    shell_choices = get_completion_shells()
    shell_arg = parser.add_argument("shell", choices=shell_choices, help=_("args.show.completion.shell"))

    for choice in shell_choices:
        add_completion(shell_arg, choice, _(f"args.show.completion.shell.comp.{choice}"))


def new_help_formatter_class(max_help_position: int) -> Type[HelpFormatter]:

    class CustomHelpFormatter(RawDescriptionHelpFormatter):
        def __init__(self, prog):
            super().__init__(prog, max_help_position=max_help_position)

    return CustomHelpFormatter


def get_outputs() -> List[str]:
    return ["human-color", "human", "machine"]

def get_search_kinds() -> List[str]:
    return ["mojang", "local", "forge", "fabric", "quilt", "legacyfabric", "babric"]

def get_auth_services() -> List[str]:
    return ["microsoft", "yggdrasil"]

def get_completion_shells() -> List[str]:
    return ["bash", "zsh"]


def type_path(s: str) -> Path:
    return Path(s)

def type_path_dir(s: str) -> Path:
    return Path(s)

def type_resolution(s: str) -> Tuple[int, int]:
    parts = s.split("x")
    if len(parts) == 2:
        return (int(parts[0]), int(parts[1]))
    else:
        raise ArgumentTypeError(_("args.start.resolution.invalid", given=s))

def type_email_or_username(s: str) -> str:
    return s

def type_host(s: str) -> str:
    return s


def add_completion(action: Action, name: str, description: str):
    """Add a completion for this action, this is used by 'complete' module.
    """
    if not hasattr(action, "_pmc_completions"):
        action._pmc_completions = {} # type: ignore
    action._pmc_completions[name] = description # type: ignore

def get_completions(action: Action) -> Dict[str, str]:
    return getattr(action, "_pmc_completions", {})



"""Definition of tasks for installing and running Fabric/Quilt mod loader.
"""

from .standard import Context, VersionHandle, Version, Watcher, VersionNotFoundError
from .http import http_request, HttpError

from typing import Optional, Any, Iterator


class _FabricApiLoader:
    """This class describes a loader returned from the fabric API. (unstable API)
    """
    __slots__ = "version", "stable"
    def __init__(self, version: str, stable: bool) -> None:
        self.version = version
        self.stable = stable


class FabricApi:
    """This class is internally used to defined two constant for both official Fabric
    backend API and Quilt API which have the same endpoints. So we use the same logic
    for both mod loaders.
    """

    def __init__(self, name: str, api_url: str) -> None:
        self.name = name
        self.api_url = api_url

    def request_fabric_meta(self, method: str) -> Any:
        """Generic HTTP request to the fabric's REST API.
        """
        return http_request("GET", f"{self.api_url}{method}", accept="application/json").json()

    def request_version_loader_profile(self, vanilla_version: str, loader_version: str) -> dict:
        """Return the version profile for the given vanilla version and loader.
        """
        return self.request_fabric_meta(f"versions/loader/{vanilla_version}/{loader_version}/profile/json")

    def _request_loaders(self, vanilla_version: Optional[str] = None) -> Iterator[_FabricApiLoader]:
        """Return an iterator of loaders available for the given vanilla version, if no
        vanilla version is specified, this returned an iterator of all loaders.
        """
        
        def map_loader(obj) -> _FabricApiLoader:
            return _FabricApiLoader(str(obj.get("version", "")), bool(obj.get("stable", False)))

        if vanilla_version is not None:
            loaders = self.request_fabric_meta(f"versions/loader/{vanilla_version}")
            return map(lambda obj: map_loader(obj["loader"]), loaders)
        else:
            return map(map_loader, self.request_fabric_meta("versions/loader"))

    def _request_latest_loader(self, vanilla_version: Optional[str] = None) -> Optional[_FabricApiLoader]:
        """Return the latest loader version for the given vanilla version, if no vanilla
        version is specified, this return the latest loader.
        """
        try:
            return next(self._request_loaders(vanilla_version))
        except StopIteration:
            return None

    # DEPRECATED:
    
    def request_fabric_loader_versions(self) -> Iterator[str]:
        """ deprecated, will be replaced by request_loaders """
        return map(lambda loader: loader.version, self._request_loaders())

    def request_fabric_loader_version(self, vanilla_version: str) -> Optional[str]:
        """ deprecated, will be replaced by request_latest_loader """
        loader = self._request_latest_loader(vanilla_version)
        return None if loader is None else loader.version


FABRIC_API = FabricApi("fabric", "https://meta.fabricmc.net/v2/")
QUILT_API = FabricApi("quilt", "https://meta.quiltmc.org/v3/")
LEGACYFABRIC_API = FabricApi("legacyfabric", "https://meta.legacyfabric.net/v2/")
BABRIC_API = FabricApi("babric", "https://meta.babric.glass-launcher.net/v2/")


class FabricVersion(Version):

    def __init__(self, api: FabricApi, vanilla_version: str, loader_version: Optional[str],
        prefix: str, *,
        context: Optional[Context] = None,
    ) -> None:

        super().__init__("", context=context)  # Do not give a root version for now.

        self.api = api
        self.vanilla_version = vanilla_version
        self.loader_version = loader_version
        self.prefix = prefix

    @classmethod
    def with_fabric(cls, vanilla_version: str = "release", loader_version: Optional[str] = None, *,
        context: Optional[Context] = None,
        prefix: str = "fabric"
    ) -> "FabricVersion":
        """Construct a root for resolving a Fabric version.
        """
        return cls(FABRIC_API, vanilla_version, loader_version, prefix, context=context)

    @classmethod
    def with_quilt(cls, vanilla_version: str = "release", loader_version: Optional[str] = None, *,
        context: Optional[Context] = None,
        prefix: str = "quilt"
    ) -> "FabricVersion":
        """Construct a root for resolving a Quilt version.
        """
        return cls(QUILT_API, vanilla_version, loader_version, prefix, context=context)

    @classmethod
    def _with_legacyfabric(cls, vanilla_version: str = "release", loader_version: Optional[str] = None, *,
        context: Optional[Context] = None,
        prefix="legacyfabric"
    ) -> "FabricVersion":
        """Construct a root for resolving a LegacyFabric version"""
        return cls(LEGACYFABRIC_API, vanilla_version, loader_version, prefix, context=context)

    @classmethod
    def _with_babric(cls, vanilla_version: str = "release", loader_version: Optional[str] = None, *,
        context: Optional[Context] = None,
        prefix="babric"
    ) -> "FabricVersion":
        """Construct a root for resolving a LegacyFabric version"""
        return cls(BABRIC_API, vanilla_version, loader_version, prefix, context=context)

    def _resolve_version(self, watcher: Watcher) -> None:
        
        # Vanilla version may be "release" or "snapshot"
        self.vanilla_version = self.manifest.filter_latest(self.vanilla_version)[0]

        # Resolve loader version if not specified.
        if self.loader_version is None:

            watcher.handle(FabricResolveEvent(self.api, self.vanilla_version, None))

            try:
                loader = self.api._request_latest_loader(self.vanilla_version)
                self.loader_version = None if loader is None else loader.version
            except HttpError as error:
                if error.res.status not in (404, 400):
                    raise
                self.loader_version = None

            if self.loader_version is None:
                # Correct error if the error is just a not found.
                raise VersionNotFoundError(f"{self.prefix}-{self.vanilla_version}-???")

            watcher.handle(FabricResolveEvent(self.api, self.vanilla_version, self.loader_version))

        # Finally define the full version id.
        self.version = f"{self.prefix}-{self.vanilla_version}-{self.loader_version}"

    def _load_version(self, version: VersionHandle, watcher: Watcher) -> bool:
        if version.id == self.version:
            return version.read_metadata_file()
        else:
            return super()._load_version(version, watcher)

    def _fetch_version(self, version: VersionHandle, watcher: Watcher) -> None:

        if version.id != self.version:
            return super()._fetch_version(version, watcher)

        assert self.loader_version is not None, "_resolve_fabric_loader(...) missing"

        try:
            version.metadata = self.api.request_version_loader_profile(self.vanilla_version, self.loader_version)
        except HttpError as error:
            if error.res.status not in (404, 400):
                raise
            # Correct error if the error is just a not found.
            raise VersionNotFoundError(version.id)

        version.metadata["id"] = version.id
        version.write_metadata_file()


class FabricResolveEvent:
    """Event triggered when the loader version is missing and is being resolved.
    """
    __slots__ = "api", "vanilla_version", "loader_version"
    def __init__(self, api: FabricApi, vanilla_version: str, loader_version: Optional[str]) -> None:
        self.api = api
        self.vanilla_version = vanilla_version
        self.loader_version = loader_version



"""CLI languages management.
"""

from portablemc.standard import JvmNotFoundError, JvmLoadedEvent, Version
from portablemc.forge import ForgeInstallError
from portablemc.util import jvm_bin_filename
from portablemc.download import DownloadResultError

from typing import Optional


def get_raw(key: str, kwargs: Optional[dict], default: Optional[str] = None) -> str:
    """Get a message translated using the given keyword formatting arguments.

    :param key: The key of the message to translate.
    :param kwargs: The keyword formatting dictionary.
    :return: Translated message, or the default value if not found. By default, the 
    default value if the key itself.
    """
    try:
        return lang[key].format_map(kwargs or {})
    except KeyError:
        return key


def get(key: str, **kwargs) -> str:
    """Get a message translated using the given keyword formatting arguments.

    :param key: The key of the message to translate.
    :return: Translated message, or the key itself if not found.
    """
    return get_raw(key, kwargs)


lang = {
    # Args root
    "args._": 
        "  A fast, reliable and cross-platform command-line Minecraft launcher and API\n"
        "  for developers. Including fast and easy installation of common mod loaders such\n"
        "  as Fabric, LegacyFabric, Forge, NeoForge, Quilt and Babric. This launcher is\n" 
        "  compatible with the standard Minecraft directories.\n\n",
    "args.main_dir": "Set the main directory where libraries, assets and versions.",
    "args.work_dir": "Set the working directory where the game run and place for examples "
        "saves, screenshots (and resources for legacy versions), it also store "
        "runtime binaries and authentication.",
    "args.timeout": "Set a global timeout (in decimal seconds) for network requests.",
    "args.output": "Set the output format of the launcher, defaults to human-color, human if not a TTY.",
    "args.output.comp.human-color": "Human readable output with color.",
    "args.output.comp.human": "Human readable output.",
    "args.output.comp.machine": "Machine readable output.",
    "args.verbose": "Enable verbose output. The more -v argument you put, the more verbose "
        "the launcher will be, depending on subcommands' support (usually -v, -vv, -vvv).",
    # Args common langs
    "args.common.help": "Show this help message and exit.",
    "args.common.auth_service": "Authentication service type to use for logging in the game.",
    "args.common.auth_service.comp.microsoft": "Microsoft authentication (default).",
    "args.common.auth_service.comp.yggdrasil": "Mojang authentication (deprecated).",
    "args.common.auth_no_browser": "Prevent the authentication service to open your system's web browser.",
    # Args search
    "args.search": "Search for versions.",
    "args.search._": 
        "  Search for versions, by default this command will search for official Mojang version\n"
        "  but you can change this behavior and search for local or mod loaders versions with the\n"
        "  -k (--kind) argument. Note that the displayed table layout depends on the version kind.\n"
        "  There is a special case when using version aliases 'release' or 'snapshot', in such case\n"
        "  the version alias is resolved and the real version is displayed. If no filter is given,\n"
        "  all results are displayed.\n\n"
        "    $ portablemc search\n"
        "    $ portablemc search release\n",
    "args.search.kind": "Select the kind of search to operate.",
    "args.search.kind.comp.mojang": "Search for official Mojang versions (default).",
    "args.search.kind.comp.local": "Search for locally installed versions.",
    "args.search.kind.comp.forge": "Search for Forge versions.",
    "args.search.kind.comp.fabric": "Search for Fabric versions.",
    "args.search.kind.comp.legacyfabric": "Search for LegacyFabric versions.",
    "args.search.kind.comp.quilt": "Search for Quilt versions.",
    "args.search.kind.comp.babric": "Search for Babric versions.",
    "args.search.input": "Search input.",
    "args.search.input.comp.release": "Resolve version of the latest release.",
    "args.search.input.comp.snapshot": "Resolve version of the latest snapshot.",
    # Args start
    "args.start": "Start the game.",
    "args.start.version": "Version identifier (default to release): {formats}.",
    "args.start.version.standard": "release|snapshot|<vanilla-version>",
    "args.start.version.fabric": "fabric:[<vanilla-version>[:<loader-version>]]",
    "args.start.version.legacyfabric": "legacyfabric:[<vanilla-version>[:<loader-version>]]",
    "args.start.version.quilt": "quilt:[<vanilla-version>[:<loader-version>]]",
    "args.start.version.babric": "babric::[<loader-version>]",
    "args.start.version.forge": "forge:[<forge-version>] (forge-version >= 1.5.2)",
    "args.start.version.neoforge": "neoforge:[<neoforge-version>] (neoforge-version >= 1.20.1)",
    "args.start.version.comp.release": "Start the latest release (default).",
    "args.start.version.comp.snapshot": "Start the latest snapshot.",
    "args.start.version.comp.fabric": "Start Fabric mod loader with latest release.",
    "args.start.version.comp.legacyfabric": "Start LegacyFabric mod loader with latest release.",
    "args.start.version.comp.quilt": "Start Quilt mod loader with latest release.",
    "args.start.version.comp.babric": "Start Babric mod loader with beta 1.7.3.",
    "args.start.version.comp.forge": "Start Forge mod loader with latest release.",
    "args.start.version.comp.neoforge": "Start NeoForge mod loader with latest release.",
    "args.start.dry": "Simulate game starting.",
    "args.start.disable_multiplayer": "Disable the multiplayer buttons (>= 1.16).",
    "args.start.disable_chat": "Disable the online chat (>= 1.16).",
    "args.start.demo": "Start game in demo mode.",
    "args.start.resolution": "Set a custom start resolution (<width>x<height>, >= 1.6).",
    "args.start.resolution.invalid": "invalid format '{given}', expected <width>x<height>",
    "args.start.jvm": f"Set a custom JVM '{jvm_bin_filename}' executable path. If this argument is omitted a public build "
        "of a JVM is downloaded from Mojang services (if Mojang does not support your system, error is returned).",
    "args.start.jvm_args": "Change the default JVM arguments.",
    "args.start.no_fix": "Flag that globally disable fixes (proxy for old versions), "
        "enabled by default.",
    "args.start.fabric_prefix": "Change the prefix of the version ID when starting with Fabric (<prefix>-<vanilla-version>-<loader-version>).",
    "args.start.legacyfabric_prefix": "Change the prefix of the version ID when starting with LegacyFabric (<prefix>-<vanilla-version>-<loader-version>).",
    "args.start.babric_prefix": "Change the prefix of the version ID when starting with Babric (<prefix>-<vanilla-version>-<loader-version>).",
    "args.start.quilt_prefix": "Change the prefix of the version ID when starting with Quilt (<prefix>-<vanilla-version>-<loader-version>).",
    "args.start.forge_prefix": "Change the prefix of the version ID when starting with Forge (<prefix>-<forge-version>).",
    "args.start.neoforge_prefix": "Change the prefix of the version ID when starting with NeoForge (<prefix>-<neoforge-version>).",
    "args.start.lwjgl": "Change the default LWJGL version used by Minecraft (LWJGL >= 3.2.3). "
        "This argument makes additional changes in order to support processor architectures such as ARM. "
        "It's not guaranteed to work with every version of Minecraft and downgrading LWJGL version is not recommended.",
    "args.start.exclude_lib": "Specify Java libraries to exclude from the classpath (and download) "
        "before launching the game. Follow this pattern to specify libraries: <artifact>[:[<version>][:<classifier>]]. "
        "If your system doesn't support Mojang-provided natives, you can use both --exclude-lib and "
        "--include-bin to replace them with your own (e.g. --exclude-lib lwjgl-glfw::natives --include-bin /lib/libglfw.so).",
    "args.start.include_bin": "Include binaries (.so, .dll, .dylib) in the bin directory of the game, "
        "given files are symlinked in the directory if possible, copied if not. "
        "On linux, version numbers are discarded (e.g. /usr/lib/foo.so.1.22.2 -> foo.so). "
        "Read the --exclude-lib help for use cases.",
    "args.start.auth_anonymize": "Anonymize your email or username for authentication messages.",
    "args.start.temp_login": "Flag used with -l (--login) to tell launcher not to cache your session if "
        "not already cached, disabled by default.",
    "args.start.login": "Use a email (or deprecated username) to authenticate using selected "
        "service (with --auth-service, also overrides --username and --uuid).",
    "args.start.username": "Set a custom user name to play.",
    "args.start.uuid": "Set a custom user UUID to play.",
    "args.start.server": "Start the game and directly connect to a multiplayer server (>= 1.6).",
    "args.start.server_port": "Set the server port (given with -s, --server, >= 1.6).",
    # Args login
    "args.login": "Login into your account and save the session.",
    # Args logout
    "args.logout": "Logout and invalidate a session.",
    # Args show
    "args.show": "Show, debug and generate data unrelated to the game.",
    "args.show.about": "Display authors, version and license of PortableMC.",
    "args.show.auth": "Debug the authentication database and supported services.",
    "args.show.lang": "Debug the language mappings used for messages translation.",
    "args.show.completion": "Print a shell completion script.",
    "args.show.completion._": 
        # Part of this description are from 'rustup' completion description.
        "  This command prints a shell completion script in the terminal.\n"
        "  The installation of this completion script depends on you shell and is explained below.\n\n"
        "  BASH:\n\n"
        "  Completion files are commonly stored in '/etc/bash_completion.d/' for system-wide commands,\n"
        "  but can be stored in '~/.local/share/bash-completion/completions' for user-specific commands.\n"
        "  You can run the following commands to generate the file:\n\n"
        "    $ mkdir -p ~/.local/share/bash-completion/completions\n"
        "    $ portablemc show completion bash > ~/.local/share/bash-completion/completions/portablemc\n\n"
        "  You can also dynamically evaluate the script, but it may slow your shell startup:\n\n"
        "    $ eval \"$(portablemc show completion bash)\"\n\n"
        "  ZSH:\n\n"
        "  Zsh completions are commonly stored in any directory listed in your '$fpath' variable.\n"
        "  To use these completions, you must either add the generated script to one of those\n"
        "  directories, or add your own to this list. Once you chose a '$fpath' directory:\n\n"
        "    $ portablemc show completion zsh > your-dir/_portablemc\n\n"
        "  You can also dynamically evaluate a script, but it may slow your shell startup:\n\n"
        "    $ eval \"$(portablemc show completion zsh)\"\n\n",
    "args.show.completion.shell": "The shell to generate completion script for (default to your current shell, required if not found).",
    "args.show.completion.shell.comp.bash": "Generate completion script for Bash.",
    "args.show.completion.shell.comp.zsh": "Generate completion script for Zsh.",
    # Common
    "echo": "{echo}",
    "cancelled": "Cancelled.",
    "keyboard_interrupt": "Keyboard interrupted.",
    "suggest_verbose": "Use verbose flag -v to get informations that may be useful for developers.",
    # Common errors
    "error.os": "An unexpected OS error happened:",
    "error.http": "Unhandled HTTP error happened:",
    "error.socket": "This operation requires an operational network, but a socket error happened:",
    "error.socket.tip.version_manifest": "Version manifest may not be locally cached, try to run this command once with an operational network.",
    "error.socket.tip.fabric_loader_version": "Fabric loader version must be specified if network is not operational.",
    "error.socket.tip.legacyfabric_loader_version": "Fabric loader version must be specified if network is not operational.",
    "error.socket.tip.quilt_loader_version": "Quilt loader version must be specified if network is not operational.",
    "error.socket.tip.babric_loader_version": "Babric loader version must be specified if network is not operational.",
    "error.cert": "Certificate verification failed, you can try installing 'certifi' package:",
    # Command search
    "search.type": "Type",
    "search.name": "Identifier",
    "search.release_date": "Release date",
    "search.last_modified": "Last modified",
    "search.flags": "Flags",
    "search.flags.local": "local",
    "search.flags.stable": "stable",
    "search.loader_version": "Loader version",
    # Command login
    "login.tip.remember_start_login": "Remember to start the game with '-l {email}' if you want to be authenticated in-game.",
    # Command logout
    "logout.yggdrasil.pending": "Logging out {email} from Mojang...",
    "logout.microsoft.pending": "Logging out {email} from Microsoft...",
    "logout.success": "Logged out {email}",
    "logout.unknown_session": "No session for {email}",
    # Command start
    "start.global_version": "Global version: {kind} {version} {remaining}",
    "start.version.invalid_id": "Invalid version id, expected: {expected}",
    "start.version.invalid_id_unknown_kind": "Invalid version id, unknown kind: {kind}.",
    "start.version.loading": "Loading version {version}... ",
    "start.version.fetching": "Fetching version {version}... ",
    "start.version.loaded": "Loaded version {version}",
    "start.version.loaded.fetched": "Loaded version {version} (fetched)",
    "start.version.not_found": "Version {version} not found",
    "start.version.too_much_parents": "Too much parents while resolving versions.",
    "start.features": "Features: [{features}]",
    "start.jar.found": "Checked version jar",
    "start.jar.not_found": "Version jar not found",
    "start.assets.resolving": "Checking assets version {index_version}... ",
    "start.assets.resolved": "Checked {count} assets version {index_version}",
    "start.libraries.resolving": "Checking libraries...",
    "start.libraries.resolved": "Checked {class_libs_count} class and {native_libs_count} native libraries",
    "start.libraries.excluded": "Excluded library {spec}",
    "start.libraries.unused_filter": "Unused library filter {filter}",
    "start.libraries.not_found_error": "A required library is not installed but no way to download it: {spec}",
    "start.logger.found": "Using logger {version}",
    "start.jvm.loading": "Loading java...",
    f"start.jvm.loaded.{JvmLoadedEvent.MOJANG}": "Loaded Mojang java {version}",
    f"start.jvm.loaded.{JvmLoadedEvent.BUILTIN}": "Loaded builtin java {version}",
    f"start.jvm.loaded.{JvmLoadedEvent.CUSTOM}": "Loaded custom java {version}",
    f"start.jvm.not_found_error.{JvmNotFoundError.UNSUPPORTED_ARCH}": "No JVM download was found for your platform architecture, "
        "use --jvm argument to manually set the path to your JVM executable.",
    f"start.jvm.not_found_error.{JvmNotFoundError.UNSUPPORTED_VERSION}": "No JVM download was found, "
        "use --jvm argument to manually set the path to your JVM executable.",
    f"start.jvm.not_found_error.{JvmNotFoundError.UNSUPPORTED_LIBC}": "No JVM download was found for your libc (only glibc is supported), "
        "use --jvm argument to manually set the path to your JVM executable.",
    f"start.jvm.not_found_error.{JvmNotFoundError.BUILTIN_INVALID_VERSION}": f"The builtin JVM ({jvm_bin_filename}) is not compatible "
        "with selected game version.",
    f"start.fix.{Version.FIX_LEGACY_PROXY}": "Using legacy proxy for online resources: {value}",
    f"start.fix.{Version.FIX_LEGACY_MERGE_SORT}": "Using legacy merge sort: {value}",
    f"start.fix.{Version.FIX_LEGACY_RESOLUTION}": "Included resolution into game arguments: {value}",
    f"start.fix.{Version.FIX_LEGACY_QUICK_PLAY}": "Included legacy quick play into game arguments: {value}",
    f"start.fix.{Version.FIX_AUTH_LIB_2_1_28}": "Fixed authlib for 1.16.4 or 1.16.5: {value}",
    f"start.fix.{Version.FIX_LWJGL}": "Fixed LWJGL: {value}",
    "start.additional_binary_not_found": "The additional binary '{path}' doesn't exists.",
    "start.bin_install": "Installed binary {src_file} as {dst_name}",
    # Command start (LWJGL fix)
    "start.lwjgl.version": "Forced LWJGL version to {version}",
    # Command start (fabric)
    "start.fabric.resolving": "Resolving {api} loader for {vanilla_version}...",
    "start.fabric.resolved": "Resolved {api} loader {loader_version} for {vanilla_version}",
    # Command start (forge)
    "start.forge.resolving": "Resolving {api} alias {version}...",
    "start.forge.resolved": "Resolved {api} {version}",
    "start.forge.post_processing": "Forge post processing: {task}...",
    "start.forge.post_processed": "Forge post processing done",
    f"start.forge.install_error.{ForgeInstallError.INSTALL_PROFILE_NOT_FOUND}": "Install profile not found in the forge installer.",
    f"start.forge.install_error.{ForgeInstallError.VERSION_METADATA_NOT_FOUND}": "Version metadata not found in the forge installer.",
    # Pretty download
    "download.threads_count": "Download threads count: {count}",
    "download.start": "Download starting...",
    "download.progress": "Download: {count}/{total_count} {size:>8} @ {speed}",
    "download.error": "{name}: {message}",
    f"download.error.{DownloadResultError.CONNECTION}": "Connection error",
    f"download.error.{DownloadResultError.NOT_FOUND}": "Not found",
    f"download.error.{DownloadResultError.INVALID_SIZE}": "Invalid size",
    f"download.error.{DownloadResultError.INVALID_SHA1}": "Invalid SHA1",
    # Auth common
    "auth.refreshing": "Invalid session, refreshing...",
    "auth.refreshed": "Session refreshed for {email}",
    "auth.validated": "Session validated for {email}",
    "auth.caching": "Caching your session...",
    "auth.logged_in": "Session logged for {email}",
    "auth.error": "Error authenticating: {message}",
    # Auth Yggdrasil
    "auth.yggdrasil": "Authenticating {email} with Mojang...",
    "auth.yggdrasil.enter_password": "Password: ",
    "auth.yggdrasil.deprecated": "Mojang authentication is deprecated and does not work anymore.",
    # Auth Microsoft
    "auth.microsoft": "Authenticating {email} with Microsoft...",
    "auth.microsoft.no_browser_fallback": "Authenticating without local browser, please go to the following url to login:",
    "auth.microsoft.no_browser_code": "Paste the code: ",
    "auth.microsoft.opening_browser_and_listening": "Opened authentication page in browser...",
    "auth.microsoft.close_tab_and_return": "Close this tab and return to the launcher.",
    "auth.microsoft.failed_to_authenticate": "Failed to authenticate.",
    "auth.microsoft.processing": "Processing authentication against Minecraft services...",
    "auth.microsoft.incoherent_data": "Incoherent authentication data, please retry.",
}



"""Global utilities used internally. The functions can be used externally but upward
compatibility is not guaranteed unless explicitly specified.
"""

from datetime import datetime
import platform

from typing import Optional


jvm_bin_filename = "javaw.exe" if platform.system() == "Windows" else "java"


def merge_dict(dst: dict, other: dict) -> None:
    """Merge a dictionary into a destination one.

    Merge the `other` dict into the `dst` dict. For every key/value in `other`, if the key
    is present in `dst`it does nothing. Unless values in both dict are also dict, in this
    case the merge is recursive. If the value in both dict are list, the 'dst' list is 
    extended (.extend()) with the one of `other`. If a key is present in both `dst` and
    `other` but with different types, the value is not overwritten.

    :param dst: The source dictionary to merge `other` into.
    :param other: The dictionary merged into `dst`.
    """

    for k, v in other.items():
        if k in dst:
            dst_v = dst[k]
            if isinstance(dst_v, dict) and isinstance(v, dict):
                merge_dict(dst_v, v)
            elif isinstance(dst_v, list) and isinstance(v, list):
                dst[k] = v + dst_v
        else:
            dst[k] = v


def calc_input_sha1(input_stream, *, buffer_len: int = 8192) -> str:
    """Internal function to calculate the sha1 of an input stream.

    :param input_stream: The input stream that supports `readinto`.
    :param buffer_len: Internal buffer length, defaults to 8192
    :return: The sha1 string.
    """
    import hashlib
    h = hashlib.sha1()
    b = bytearray(buffer_len)
    mv = memoryview(b)
    for n in iter(lambda: input_stream.readinto(mv), 0):
        h.update(mv[:n])
    return h.hexdigest()


def from_iso_date(raw: str) -> datetime:
    """Replacement for `datetime.fromisoformat()` which is missing from Python 3.6. This 
    function replace it if needed.

    Currently, only a subset of the ISO format is supported, both hours, minutes and 
    seconds must be defined and the timezone, if present must contain both hours and 
    minutes, no more.
    """
    if hasattr(datetime, "fromisoformat"):
        return datetime.fromisoformat(raw)
    from datetime import timezone, timedelta
    tz_idx = raw.find("+")
    dt = datetime.strptime(raw[:tz_idx], "%Y-%m-%dT%H:%M:%S")
    if tz_idx != -1:
        tz_dt = datetime.strptime(raw[tz_idx + 1:], "%H:%M")
        dt = dt.replace(tzinfo=timezone(timedelta(hours=tz_dt.hour, minutes=tz_dt.minute)))
    return dt


class LibrarySpecifier:
    """A maven-style library specifier.
    """

    __slots__ = "group", "artifact", "version", "classifier", "extension"

    def __init__(self, group: str, artifact: str, version: str, classifier: Optional[str] = None, extension: str = "jar"):
        self.group = group
        self.artifact = artifact
        self.version = version
        self.classifier = classifier
        self.extension = extension
    
    @classmethod
    def from_str(cls, s: str) -> "LibrarySpecifier":
        """Parse a library specifier string 'group:artifact:version[:classifier]'.
        """

        ext_split = s.rsplit("@", maxsplit=1)
        ext = "jar" if len(ext_split) == 1 else ext_split[1]

        if not len(ext):
            raise ValueError("invalid library specifier: empty extension")
        
        parts = ext_split[0].split(":", 3)

        if len(parts) < 3:
            raise ValueError("invalid library specifier: too few parts")
        else:
            return LibrarySpecifier(parts[0], parts[1], parts[2], parts[3] if len(parts) == 4 else None, ext)

    def __str__(self) -> str:
        return f"{self.group}:{self.artifact}:{self.version}" + \
            ("" if self.classifier is None else f":{self.classifier}") + \
            ("" if self.extension == "jar" else f"@{self.extension}")

    def __eq__(self, other) -> bool:
        return isinstance(other, LibrarySpecifier) and \
            (self.group, self.artifact, self.version, self.classifier, self.extension) == \
            (other.group, other.artifact, other.version, other.classifier, other.extension)

    def __repr__(self) -> str:
        return f"<LibrarySpecifier {self}>"
    
    def __hash__(self) -> int:
        return hash((self.group, self.artifact, self.version, self.classifier, self.extension))

    def file_path(self) -> str:
        """Return the standard path to store the file of this specifier.
        
        The path separator will always be forward slashes '/', because it's compatible 
        with linux/mac/windows and URL paths.

        Specifier `com.foo.bar:artifact:version@zip` gives 
        `com/foo/bar/artifact/version/artifact-version.zip`.
        """
        
        file_name = f"{self.artifact}-{self.version}" + \
            ("" if self.classifier is None else f"-{self.classifier}") + \
            f".{self.extension}"
        
        return "/".join([*self.group.split("."), self.artifact, self.version, file_name])

    def copy(self) -> "LibrarySpecifier":
        return LibrarySpecifier(self.group, self.artifact, self.version, self.classifier, self.extension)



"""HTTP primitive functions.
"""

from urllib.error import HTTPError, URLError
from http.client import HTTPResponse
import urllib.request
import urllib.parse
import json
import ssl

from . import LAUNCHER_VERSION

from typing import Optional, Any, cast


__all__ = ["HttpResponse", "HttpError", "http_request"]


class HttpResponse:
    """An HTTP response containing the status, data and received headers.
    """
    
    def __init__(self, res: Optional[HTTPResponse]) -> None:

        self.status = 0 if res is None else res.status
        self.data = b"null" if res is None else res.read()
        self.headers = {}

        if res is not None:
            for header_name, header_value in res.getheaders():
                self.headers[header_name] = header_value

    def json(self) -> Any:
        """Parse the data as JSON. This may raise a JSONDecodeError.
        """
        return json.loads(self.data)
    
    def text(self) -> str:
        """Parse the data as UTF-8 text.
        """
        return self.data.decode()

    def __repr__(self) -> str:
        return f"<HttpResponse {self.status}>"


class HttpError(Exception):
    """An HTTP error, raised when the status code of the response is not 200.

    If any network error happens and it's impossible to receive a response from the 
    server, an instance of `HttpResponse` with status equal to 0 is used (also has no 
    headers and `None` data). The original reason for this error is given in the `reason`
    attribute in any case.
    """

    def __init__(self, res: HttpResponse, method: str, url: str, reason: URLError) -> None:
        self.res = res
        self.method = method
        self.url = url
        self.reason = reason

    def __repr__(self) -> str:
        return f"<HttpError {self.res}, origin: {self.method} {self.url}, reason: {self.reason}>"


def http_request(method: str, url: str, *,
    data: Optional[bytes] = None,
    headers: Optional[dict] = None,
    accept: Optional[str] = None,
    content_type: Optional[str] = None
) -> HttpResponse:
    """Make a synchronous HTTP request.

    :return: The response returned should've a status of 2xx.
    :raises HttpError: An error wrapping a response that is not of status 2xx.
    """
    
    if headers is None:
        headers = {}
    if accept is not None:
        headers["Accept"] = accept
    if content_type is not None:
        headers["Content-Type"] = content_type
    if "User-Agent" not in headers:
        headers["User-Agent"] = f"portablemc/{LAUNCHER_VERSION}"

    try:
        import certifi
        ctx = ssl.create_default_context(cafile=certifi.where())
    except ImportError:
        ctx = None

    try:
        req = urllib.request.Request(url, data, headers, method=method)
        res: HTTPResponse = urllib.request.urlopen(req, context=ctx)
        return HttpResponse(res)
    except HTTPError as error:
        raise HttpError(HttpResponse(cast(HTTPResponse, error)), method, url, error)
    except URLError as error:
        raise HttpError(HttpResponse(None), method, url, error)



"""Mojang and Microsoft authentication utilities.
"""

from urllib import parse as url_parse
from uuid import UUID, uuid4, uuid5
from pathlib import Path
import platform
import base64
import json

from .http import HttpError, http_request

from typing import Optional, Dict, Type, Tuple


class AuthSession:
    """An abstract class for defining authentication sessions. These sessions are then
    provided as an argument for starting the game. They provide all information such as
    access player's token, username or UUID.

    Different class variables must be defined by subclasses: `db_type` which is used when
    saving the session to the database, `user_type` is an information sent through command
    line to the game and `fields` which defines all class's object's fields that should
    be saved and restored from the database.

    As stated above, the session class is closely related to `AuthDatabase` in which they
    are saved and later restored.
    """

    db_type: str
    user_type: str
    fields = "access_token", "username", "uuid", "client_id"

    @classmethod
    def fix_data(cls, data: dict) -> None:
        """This optional function may be used by subclass to provide data migration from
        older database formats. The input data can be modified as needed to fit currently
        required fields (specified in `fields` class variable).

        :param data: The data that can be modified if relevant.
        """

    def __init__(self):
        self.access_token = ""
        self.username = ""
        self.uuid = ""
        self.client_id = ""

    def format_token_argument(self, legacy: bool) -> str:
        """Format the token for the game's command line. Modern versions uses the format
        `token:{access_token}:{uuid}` and legacy versions uses `{access_token}`.

        :param legacy: True to enable legacy formatting, used by older versions.
        :return: The formatted token.
        """
        return f"token:{self.access_token}:{self.uuid}" if legacy else self.access_token

    def get_xuid(self) -> str:
        """Getter specific to Microsoft, but common to auth sessions because it's used for
        Minecraft's command line arguments.
        """
        return ""

    def validate(self) -> bool:
        """Validate that the current session is still actually authenticating the player.

        :return: True if the session is still valid, when returning false the `refresh`
        method may be called to update the session.
        """
        return True

    def refresh(self) -> None:
        """Try refreshing the session to make it valid again.
        """

    def invalidate(self) -> None:
        """Invalid the session so that `validate` will now return false.
        """


class OfflineAuthSession(AuthSession):
    """Offline session, this is quite contradictory but it's actually useful to simplify
    the start logic. It provides optional static username and UUID and random when kept
    unspecified.
    """

    db_type = "offline"
    user_type = ""

    def __init__(self, username: Optional[str], uuid: Optional[str]):
        super().__init__()
        if uuid is not None and len(uuid) == 32:
            # If the UUID is already valid.
            self.uuid = uuid
            self.username = uuid[:8] if username is None else username[:16]
        else:
            namespace_hash = UUID("8df5a464-38de-11ec-aa66-3fd636ee2ed7")
            if username is None:
                self.uuid = uuid5(namespace_hash, platform.node()).hex
                self.username = self.uuid[:8]
            else:
                self.username = username[:16]
                self.uuid = uuid5(namespace_hash, self.username).hex

    def format_token_argument(self, legacy: bool) -> str:
        return ""


class YggdrasilAuthSession(AuthSession):
    """Yggdrasil authentication (deprecated). This authentication is now deprecated but
    was also known as "Mojang authentication".
    """

    db_type = "yggdrasil"
    user_type = "mojang"

    @classmethod
    def fix_data(cls, data: dict):
        if "client_token" in data:
            data["client_id"] = data.pop("client_token")

    def validate(self) -> bool:
        return self.request("validate", {
            "accessToken": self.access_token,
            "clientToken": self.client_id
        }, False)[0] == 204

    def refresh(self):
        _, res = self.request("refresh", {
            "accessToken": self.access_token,
            "clientToken": self.client_id
        })
        self.access_token = res["accessToken"]
        self.username = res["selectedProfile"]["name"]  # Refresh username if renamed (does it works? to check.).

    def invalidate(self):
        self.request("invalidate", {
            "accessToken": self.access_token,
            "clientToken": self.client_id
        }, False)

    @classmethod
    def authenticate(cls, client_id: str, email: str, password: str) -> 'YggdrasilAuthSession':
        _, res = cls.request("authenticate", {
            "agent": {
                "name": "Minecraft",
                "version": 1
            },
            "username": email,
            "password": password,
            "clientToken": client_id
        })
        sess = cls()
        sess.access_token = res["accessToken"]
        sess.username = res["selectedProfile"]["name"]
        sess.uuid = res["selectedProfile"]["id"]
        sess.client_id = res["clientToken"]
        return sess

    @classmethod
    def request(cls, req: str, payload: dict, raise_error: bool = True) -> Tuple[int, dict]:
        try:
            res = http_request("POST", f"https://authserver.mojang.com/{req}", 
                data=json.dumps(payload).encode("ascii"),
                accept="application/json",
                content_type="application/json")
            return res.status, res.json()
        except HttpError as error:
            try:
                if raise_error:
                    raise AuthError(error.res.json()["errorMessage"])
                else:
                    return error.res.status, error.res.json()
            except json.JSONDecodeError:
                if raise_error:
                    raise AuthError("invalid json")
                else:
                    return error.res.status, {}


class MicrosoftAuthSession(AuthSession):
    """Microsoft authentication for Minecraft. It involves multiples endpoint from
    Mojang, MSA and XBox Live.
    """

    db_type = "microsoft"
    user_type = "msa"
    fields = "access_token", "username", "uuid", "client_id", "refresh_token", "app_id", "redirect_uri", "xuid"

    @classmethod
    def fix_data(cls, data: dict):
        if "app_id" not in data and "client_id" in data:
            data["app_id"] = data.pop("client_id")
        if "client_id" not in data or not len(data["client_id"]):
            data["client_id"] = str(uuid4())
        if "xuid" not in data:
            data["xuid"] = cls.decode_jwt_payload(data["access_token"])["xuid"]

    def __init__(self):
        super().__init__()
        self.refresh_token = ""
        self.app_id = ""
        self.redirect_uri = ""
        self.xuid = ""
        self._new_username: Optional[str] = None

    def get_xuid(self) -> str:
        return self.xuid

    def validate(self) -> bool:
        self._new_username = None
        try:
            res = self.mc_request_profile(self.access_token)
            username = res["name"]
            if self.username != username:
                self._new_username = username
                return False
            return True
        except HttpError:
            return False

    def refresh(self):
        if self._new_username is not None:
            self.username = self._new_username
            self._new_username = None
        else:
            res = self.authenticate_base({
                "client_id": self.app_id,
                "redirect_uri": self.redirect_uri,
                "refresh_token": self.refresh_token,
                "grant_type": "refresh_token",
                "scope": "xboxlive.signin"
            })
            self.access_token = res["access_token"]
            self.username = res["username"]
            self.uuid = res["uuid"]

    @staticmethod
    def get_authentication_url(app_id: str, redirect_uri: str, email: str, nonce: str):
        """deprecated"""
        return "https://login.live.com/oauth20_authorize.srf?{}".format(url_parse.urlencode({
            "client_id": app_id,
            "redirect_uri": redirect_uri,
            "response_type": "code id_token",
            "scope": "xboxlive.signin offline_access openid email",
            "login_hint": email,
            "nonce": nonce,
            "response_mode": "form_post"
        }))

    @staticmethod
    def get_logout_url(app_id: str, redirect_uri: str):
        """deprecated"""
        return "https://login.live.com/oauth20_logout.srf?{}".format(url_parse.urlencode({
            "client_id": app_id,
            "redirect_uri": redirect_uri
        }))

    @classmethod
    def check_token_id(cls, token_id: str, email: str, nonce: str) -> bool:
        """deprecated"""
        id_token_payload = cls.decode_jwt_payload(token_id)
        return id_token_payload["nonce"] == nonce and id_token_payload["email"].casefold() == email.casefold()

    @classmethod
    def authenticate(cls, client_id: str, app_id: str, code: str, redirect_uri: str) -> 'MicrosoftAuthSession':
        res = cls.authenticate_base({
            "client_id": app_id,
            "redirect_uri": redirect_uri,
            "code": code,
            "grant_type": "authorization_code",
            "scope": "xboxlive.signin"
        })
        sess = cls()
        sess.access_token = res["access_token"]
        sess.username = res["username"]
        sess.uuid = res["uuid"]
        sess.client_id = client_id
        sess.refresh_token = res["refresh_token"]
        sess.app_id = app_id
        sess.redirect_uri = redirect_uri
        sess.xuid = cls.decode_jwt_payload(res["access_token"])["xuid"]
        return sess

    @classmethod
    def authenticate_base(cls, request_token_payload: dict) -> dict:

        # Microsoft OAuth
        try:
            res = cls.ms_request("https://login.live.com/oauth20_token.srf", request_token_payload, payload_url_encoded=True)
        except HttpError as error:
            raise OutdatedTokenError(error.res.text())

        ms_refresh_token = res.get("refresh_token")

        # Xbox Live Token
        res = cls.ms_request("https://user.auth.xboxlive.com/user/authenticate", {
            "Properties": {
                "AuthMethod": "RPS",
                "SiteName": "user.auth.xboxlive.com",
                "RpsTicket": "d={}".format(res["access_token"])
            },
            "RelyingParty": "http://auth.xboxlive.com",
            "TokenType": "JWT"
        })

        xbl_token = res["Token"]
        xbl_user_hash = res["DisplayClaims"]["xui"][0]["uhs"]

        # Xbox Live XSTS Token
        res = cls.ms_request("https://xsts.auth.xboxlive.com/xsts/authorize", {
            "Properties": {
                "SandboxId": "RETAIL",
                "UserTokens": [xbl_token]
            },
            "RelyingParty": "rp://api.minecraftservices.com/",
            "TokenType": "JWT"
        })
        xsts_token = res["Token"]

        if xbl_user_hash != res["DisplayClaims"]["xui"][0]["uhs"]:
            raise AuthError("inconsistent user hash")

        # MC Services Auth
        res = cls.ms_request("https://api.minecraftservices.com/authentication/login_with_xbox", {
            "identityToken": f"XBL3.0 x={xbl_user_hash};{xsts_token}"
        })
        mc_access_token = res["access_token"]

        # MC Services Profile
        try:
            res = cls.mc_request_profile(mc_access_token)
        except HttpError as error:
            if error.res.status == 404:
                raise DoesNotOwnMinecraftError(error.res.text())
            elif error.res.status == 401:
                raise OutdatedTokenError(error.res.text())
            else:
                res = error.res.json()
                raise AuthError(res.get("errorMessage", res.get("error", "unknown error")))

        return {
            "refresh_token": ms_refresh_token,
            "access_token": mc_access_token,
            "username": res["name"],
            "uuid": res["id"]
        }

    @classmethod
    def ms_request(cls, url: str, payload: dict, *, payload_url_encoded: bool = False) -> dict:
        data = (url_parse.urlencode(payload) if payload_url_encoded else json.dumps(payload)).encode("ascii")
        content_type = "application/x-www-form-urlencoded" if payload_url_encoded else "application/json"
        return http_request("POST", url, data=data, content_type=content_type).json()

    @classmethod
    def mc_request_profile(cls, bearer: str) -> dict:
        url = "https://api.minecraftservices.com/minecraft/profile"
        return http_request("GET", url, headers={"Authorization": f"Bearer {bearer}"}).json()

    @classmethod
    def base64url_decode(cls, s: str) -> bytes:
        rem = len(s) % 4
        if rem > 0:
            s += "=" * (4 - rem)
        return base64.urlsafe_b64decode(s)

    @classmethod
    def decode_jwt_payload(cls, jwt: str) -> dict:
        return json.loads(cls.base64url_decode(jwt.split(".")[1]))


class AuthDatabase:
    """The authentication database used to keep sessions stored. It also keeps a clien
    id which is common to all sessions to identify the client.
    """

    types = {
        YggdrasilAuthSession.db_type: YggdrasilAuthSession,
        MicrosoftAuthSession.db_type: MicrosoftAuthSession
    }

    def __init__(self, file: Path):
        self.file = file
        self.sessions: Dict[str, Dict[str, AuthSession]] = {}
        self.client_id: Optional[str] = None

    def load(self):

        self.sessions.clear()

        try:
            with self.file.open("rt") as fp:
                data = json.load(fp)
                self.client_id = data.get("client_id")
                for typ, sess_type in self.types.items():
                    typ_data = data.get(typ)
                    if typ_data is not None:
                        sessions = self.sessions[typ] = {}
                        sessions_data = typ_data["sessions"]
                        for email, sess_data in sessions_data.items():
                            # Use class method fix_data to migrate data from older versions of the auth database.
                            sess_type.fix_data(sess_data)
                            sess = sess_type()
                            for field in sess_type.fields:
                                setattr(sess, field, sess_data.get(field, ""))
                            sessions[email.casefold()] = sess
        except (OSError, KeyError, TypeError, json.JSONDecodeError):
            pass

    def save(self) -> None:

        self.file.parent.mkdir(parents=True, exist_ok=True)

        with self.file.open("wt") as fp:
            data = {}
            if self.client_id is not None:
                data["client_id"] = self.client_id
            for typ, sessions in self.sessions.items():
                if typ not in self.types:
                    continue
                sess_type = self.types[typ]
                sessions_data = {}
                data[typ] = {"sessions": sessions_data}
                for email, sess in sessions.items():
                    sess_data = sessions_data[email] = {}
                    for field in sess_type.fields:
                        sess_data[field] = getattr(sess, field)
            json.dump(data, fp, indent=2)

    def get(self, email: str, sess_type: Type[AuthSession]) -> Optional[AuthSession]:
        """Try to get a session from an email and session type.
        """
        sessions = self.sessions.get(sess_type.db_type)
        return None if sessions is None else sessions.get(email.casefold())

    def put(self, email: str, sess: AuthSession):
        """Push the given authentication session to the database, updating any previous
        session with the same email for the type of session.
        """
        sessions = self.sessions.get(sess.db_type)
        if sessions is None:
            if sess.db_type not in self.types:
                raise ValueError(f"given session type '{sess.db_type}' is not supported")
            sessions = self.sessions[sess.db_type] = {}
        sessions[email.casefold()] = sess

    def remove(self, email: str, sess_type: Type[AuthSession]) -> Optional[AuthSession]:
        """Same arguments as `get` method but remove the session and return it.
        """
        email = email.casefold()
        sessions = self.sessions.get(sess_type.db_type)
        if sessions is not None:
            session = sessions.get(email)
            if session is not None:
                del sessions[email]
                return session

    def get_client_id(self) -> str:
        if self.client_id is None or len(self.client_id) != 36:
            self.client_id = str(uuid4())
        return self.client_id


class AuthError(Exception):
    pass

class DoesNotOwnMinecraftError(AuthError):
    def __init__(self, *args) -> None:
        super().__init__("does not own minecraft", *args)

class OutdatedTokenError(AuthError):
    def __init__(self, *args) -> None:
        super().__init__("outdated token", *args)



"""Main module for PortableMC API.

Please read the API documentation for more general information about the architecture of
the API and examples codes. You should also read the API documentation about 
*Versioning and stability* to understand which classes you should consider part of the
public API, and therefore stable across minor and patch version bumps.
"""

LAUNCHER_NAME = "portablemc"
LAUNCHER_VERSION = "4.4.0"
LAUNCHER_AUTHORS = ["Théo Rozier <contact@theorozier.fr>", "Contributors"]
LAUNCHER_COPYRIGHT = "PortableMC  Copyright (C) 2021-2024  Théo Rozier"
LAUNCHER_URL = "https://github.com/mindstorm38/portablemc"



from io import StringIO
from argparse import ArgumentParser, \
    _CountAction, _StoreAction, _SubParsersAction, \
    _StoreConstAction, _HelpAction, _AppendConstAction

from .lang import get as _
from .parse import type_path, type_path_dir, \
    type_email_or_username, type_host, get_completions

from typing import Dict, Tuple, cast


def gen_zsh_completion(parser: ArgumentParser) -> str:
    buffer = StringIO()
    buffer.write("#compdef portablemc\n\n")
    gen_zsh_parser_completion(parser, buffer, "_portablemc")
    buffer.write("if [[ $zsh_eval_context[-1] == loadautofunc ]]; then\n")
    buffer.write("  _portablemc\n")
    buffer.write("else\n")
    buffer.write("  compdef _portablemc portablemc\n")
    buffer.write("fi\n")
    return buffer.getvalue()

def gen_zsh_parser_completion(parser: ArgumentParser, buffer: StringIO, function: str):

    # Sources:
    # - https://zsh.sourceforge.io/Doc/Release/Completion-Widgets.html
    # - https://zsh.sourceforge.io/Doc/Release/Completion-System.html

    commands: Dict[str, Tuple[str, ArgumentParser]] = {}
    completions: Dict[str, Dict[str, str]] = {}

    buffer.write(function)
    buffer.write(" () {\n")
    buffer.write("  local curcontext=$curcontext state line\n")
    buffer.write("  integer ret=1\n")

    buffer.write("  _arguments -s -C \\\n")

    for action in parser._actions:
        
        zsh_description = escape_zsh(action.help or "")
        zsh_repeat = ""
        zsh_action = ": :"

        # Depending on the action type there are some specific things we can do.
        if isinstance(action, _CountAction):
            zsh_repeat = "\\*"
            zsh_action = ""
        elif isinstance(action, _StoreAction):

            action_completions = get_completions(action)
            if action.choices is not None:
                for choice in action.choices:
                    if choice not in action_completions:
                        action_completions[choice] = ""
            
            if action.type == type_path:
                zsh_action = ": :_files"
            elif action.type == type_path_dir:
                zsh_action = ": :_files -/"
            elif action.type == type_email_or_username:
                zsh_action = ": :_email_addresses -c"
            elif action.type == type_host:
                zsh_action = ": :_hosts"
            elif len(action_completions):
                zsh_action = f": :->arg_{action.dest}"
                completions[f"arg_{action.dest}"] = action_completions

        elif isinstance(action, (_HelpAction, _StoreConstAction, _AppendConstAction)):
            zsh_action = ""
        elif isinstance(action, _SubParsersAction):
            parsers_choices = cast(Dict[str, ArgumentParser], action.choices)
            for sub_action in action._get_subactions():
                commands[sub_action.dest] = (sub_action.help or "", parsers_choices[sub_action.dest])
            continue

        # If the argument is positional.
        if not len(action.option_strings):
            buffer.write(f"    '{zsh_action}' \\\n")
            continue
        
        # If the argument is an option.
        if len(action.option_strings) > 1:
            zsh_names = f"{{{','.join(action.option_strings)}}}"
        else:
            zsh_names = action.option_strings[0]
        buffer.write(f"    {zsh_repeat}{zsh_names}'[{zsh_description}]{zsh_action}' \\\n")
    
    if len(commands):
        buffer.write("    ': :->command' \\\n")
        buffer.write("    '*:: :->option' \\\n")

    buffer.write("    && ret=0\n")

    if len(commands) or len(completions):

        buffer.write("  case $state in\n")

        if len(commands):
            buffer.write("  command)\n")
            buffer.write("    local -a commands=(\n")
            for name, (cmd_description, cmd_parser) in commands.items():
                buffer.write(f"      '{name}:{escape_zsh(cmd_description)}'\n")
            buffer.write("    )\n")
            buffer.write("    _describe -t commands command commands && ret=0\n")
            buffer.write("    ;;\n")

            buffer.write("  option)\n")
            buffer.write("    case $line[1] in\n")
            for name, (cmd_description, cmd_parser) in commands.items():
                buffer.write(f"    {name}) {function}_{name} ;;\n")
            buffer.write("    esac\n")
            buffer.write("    ;;\n")

        for state, action_completions in completions.items():
            buffer.write(f"  {state})\n")
            buffer.write("    local -a completions=(\n")
            for name, description in action_completions.items():
                if len(description):
                    buffer.write(f"      '{escape_zsh(name)}:{escape_zsh(description)}'\n")
                else:
                    buffer.write(f"      '{escape_zsh(name)}'\n")
            buffer.write("    )\n")
            buffer.write("    _describe -t values value completions && ret=0\n")
            buffer.write("    ;;\n")
        
        buffer.write("  esac\n")

    buffer.write("}\n\n")

    for cmd_name, (cmd_description, cmd_parser) in commands.items():
        gen_zsh_parser_completion(cmd_parser, buffer, f"{function}_{cmd_name}")

def escape_zsh(s: str) -> str:
    return s.replace("'", "''").replace("[", "\\[").replace("]", "\\]").replace(":", "\\:")


def gen_bash_completion(parser: ArgumentParser) -> str:
    buffer = StringIO()
    buffer.write("#/usr/bin/env bash\n\n")
    gen_bash_parser_completion(parser, buffer, "_portablemc")
    buffer.write("\ncomplete -o filenames -o nosort -F _portablemc portablemc\n")
    return buffer.getvalue()

def gen_bash_parser_completion(parser: ArgumentParser, buffer: StringIO, function: str):

    # Note: We use single quote in this function because we double quote in bash.
    # Sources:
    # - https://www.gnu.org/software/bash/manual/html_node/Programmable-Completion-Builtins.html
    # - https://www.gnu.org/savannah-checkouts/gnu/bash/manual/bash.html
    #
    # Current limitations of the bash completer:
    # - Choices for positional arguments are not support, luckily the launcher don't use 
    #   such construct.
    # - Short arguments cannot be completed when stacked.

    buffer.write(function)
    buffer.write(' ()\n{\n')

    # We guess that there we always be two argument, the command and the argument to comp.
    buffer.write('  local index="$(( COMP_CWORD - 1 ))"\n')
    buffer.write('  local words=(${COMP_WORDS[@]:1})\n')
    buffer.write('  local word="${words[$index]}"\n')

    # Start by finding if there are sub parsers.
    commands: Dict[str, ArgumentParser] = {}
    for action in parser._actions:
        if isinstance(action, _SubParsersAction):
            commands.update(action.choices)
        elif len(action.option_strings):
            # Named argument
            buffer.write(f'  local arg_{action.dest}="{" ".join(action.option_strings)}"\n')
    
    # Write a loop to find potential sub-command, and construct arguments list.
    # We overwrite the COMP_ variables because we don't use them after loop.
    buffer.write('  for i in ${!words[@]}; do\n')

    # Start by sub-commands...
    if len(commands):
        buffer.write("    if (( i < index )); then\n")
        buffer.write("      COMP_WORDS=(${words[@]:$i})\n")
        buffer.write("      COMP_CWORD=$((index - i))\n")
        buffer.write('      case "${words[$i]}" in\n')
        for cmd_name, cmd_parser in commands.items():
            buffer.write(f'      {cmd_name}) {function}_{cmd_name}; return ;;\n')
        buffer.write("      esac\n")
        buffer.write("    fi\n")
    
    # Then arguments...
    buffer.write('    case "${words[$i]}" in\n')
    for action in parser._actions:
        if isinstance(action, (_SubParsersAction, _CountAction)):
            pass  # Count action are not limited in number
        elif len(action.option_strings):
            buffer.write( "    ")
            buffer.write( " | ".join(f'"{option}"' for option in action.option_strings))
            buffer.write(f') arg_{action.dest}="" ;;\n')
    buffer.write("    esac\n")

    buffer.write("  done\n")

    # Special case for options with associated value.
    buffer.write('  if (( index >= 1 )); then\n')
    buffer.write('    case ${words[$(( index - 1 ))]} in\n')
    
    for action in parser._actions:
        
        if isinstance(action, _StoreAction):

            if len(action.option_strings):
                
                buffer.write("    ")
                buffer.write(" | ".join(f'"{option}"' for option in action.option_strings))
                buffer.write(")\n")
                
                reply = ""

                if action.type == type_path:
                    reply = '$(compgen -o plusdirs -f -- "$word")'
                elif action.type == type_path_dir:
                    reply = '$(compgen -o plusdirs -d -- "$word")'
                elif action.type == type_email_or_username:
                    pass
                elif action.type == type_host:
                    reply = '$(compgen -A hostname -- "$word")'
                elif action.choices is not None:
                    reply = f'$(compgen -W "{" ".join(action.choices)}" -- "$word")'

                buffer.write(f"      COMPREPLY=({reply})\n")
                buffer.write( "      return\n")
                buffer.write( "      ;;\n")
            
    buffer.write("    esac\n")
    buffer.write("  fi\n")

    # This is the default reply for argument names.
    buffer.write('  COMPREPLY=($(compgen -W "')
    for cmd_name, cmd_parser in commands.items():
        buffer.write(f"{cmd_name} ")
    for action in parser._actions:
        if isinstance(action, _SubParsersAction):
            pass
        elif len(action.option_strings):
            buffer.write(f"$arg_{action.dest} ")
    buffer.write('" -- "$word"))\n')

    buffer.write('}\n\n')

    for cmd_name, cmd_parser in commands.items():
        gen_bash_parser_completion(cmd_parser, buffer, f"{function}_{cmd_name}")



"""Definition of the optimized download task.
"""

from http.client import HTTPConnection, HTTPSConnection, HTTPException
from threading import Thread
from pathlib import Path
from queue import Queue
import urllib.parse
import hashlib
import time
import ssl

from typing import Optional, Dict, List, Tuple, Union, Iterator


class DownloadEntry:
    """A download entry for the download task.
    """
    
    __slots__ = "url", "size", "sha1", "dst", "name", "executable"

    def __init__(self, 
        url: str, 
        dst: Path, *, 
        size: Optional[int] = None, 
        sha1: Optional[str] = None, 
        name: Optional[str] = None,
        executable: bool = False
    ) -> None:
        self.url = url
        self.dst = dst
        self.size = size
        self.sha1 = sha1
        self.name = url if name is None else name
        self.executable = executable

    def __repr__(self) -> str:
        return f"<DownloadEntry {self.name} @ {self.url}>"

    def __hash__(self) -> int:
        # Making size and sha1 in the hash is useful to make them, 
        # this means that once added to a dictionary, these attributes
        # should not be modified.
        return hash((self.url, self.dst, self.size, self.sha1))

    def __eq__(self, other):
        return isinstance(other, DownloadEntry) and \
            (self.url, self.dst, self.size, self.sha1) == \
            (other.url, other.dst, other.size, other.sha1)


class _DownloadEntry:
    """Internal class with already parsed URL to speed up processing and prevent 
    unsupported URL schemes.
    """

    __slots__ = "https", "host", "port", "path", "entry", "redirect"

    def __init__(self, https: bool, host: str, port: Optional[int], path: str, entry: DownloadEntry, *, redirect: int = 0) -> None:
        self.https = https
        self.host = host
        self.port = port
        self.path = path
        self.entry = entry
        self.redirect = redirect
    
    @classmethod
    def from_entry(cls, entry: DownloadEntry, *, redirect: int = 0) -> "_DownloadEntry":

        # We only support HTTP/HTTPS
        url_parsed = urllib.parse.urlparse(entry.url)
        if url_parsed.scheme not in ("http", "https"):
            raise ValueError(f"unsupported scheme '{url_parsed.scheme}://' from url {entry.url}")
        
        return cls(
            url_parsed.scheme == "https",
            url_parsed.netloc,
            url_parsed.port,
            url_parsed.path,
            entry,
            redirect=redirect)


class DownloadResult:
    """Base class for download result yielded by `DownloadList.download` function.
    """
    __slots__ = "thread_id", "entry"
    def __init__(self, thread_id: int, entry: DownloadEntry) -> None:
        self.thread_id = thread_id
        self.entry = entry


class DownloadResultProgress(DownloadResult):
    """Subclass of result when a file's download has been successful.
    """
    __slots__ = "size", "speed", "done"
    def __init__(self, thread_id: int, entry: DownloadEntry, size: int, speed: float, done: bool) -> None:
        super().__init__(thread_id, entry)
        self.size = size
        self.speed = speed
        self.done = done


class DownloadResultError(DownloadResult):
    """Subclass of result when a file's download has failed, the error code is indicated
    and the optional original error is given (for connection errors).
    """

    CONNECTION = "connection"
    TOO_MANY_REDIRECT = "too_many_redirect"
    NOT_FOUND = "not_found"
    INVALID_SIZE = "invalid_size"
    INVALID_SHA1 = "invalid_sha1"

    __slots__ = "code", "origin"

    def __init__(self, thread_id: int, entry: DownloadEntry, code: str, origin: Optional[Exception]) -> None:
        super().__init__(thread_id, entry)
        self.code = code
        self.origin = origin


class DownloadList:
    """A download list, composed of entries that can be downloaded all at once in batch
    with multithreading.
    """

    __slots__ = "entries", "count", "size", "_dst_entries"

    def __init__(self):
        self.entries: List[_DownloadEntry] = []
        self._dst_entries = {}
        self.count = 0
        self.size = 0
    
    def clear(self) -> None:
        """Clear the download entry, removing all entries and computed count/size.
        """
        self.entries.clear()
        self._dst_entries.clear()
        self.count = 0
        self.size = 0

    def add(self, entry: DownloadEntry, *, verify: bool = False) -> None:
        """Add a download entry to this list.

        :param entry: The entry to add.
        :param verify: Set to true in order to check if the file exists and has the same
        size has the given entry, in such case the entry is not added.
        """

        if entry.dst in self._dst_entries:
            raise ValueError("duplicate entry destination", entry, self._dst_entries[entry.dst])

        if verify and entry.dst.is_file() and (entry.size is None or entry.size == entry.dst.stat().st_size):
            return
        
        self.entries.append(_DownloadEntry.from_entry(entry))
        self._dst_entries[entry.dst] = entry
        self.count += 1
        if entry.size is not None:
            self.size += entry.size
    
    def download(self, threads_count: int, *,
        partial_progress: bool = False
    ) -> Iterator[Tuple[int, DownloadResult]]:
        """Execute the download.
        
        :param threads_count: The number of threads to run the download on.
        :param partial_progress: Set to true to be able to receive partial progress update
        on unfinished files, if this is false, DownloadResultProgress.done should be true.
        :return: This function returns an iterator that yields a tuple that contain the
        total number of results and the new result that came in.
        """

        # Sort our entries in order to download big files first, this is allows better
        # parallelization at start and avoid too much blocking at the end of the download.
        # Note that entries without size are considered 1 Mio, to download early.
        self.entries.sort(key=lambda e: e.entry.size or 1048576, reverse=True)

        entries_count = len(self.entries)
        if not entries_count or threads_count < 1:
            return

        threads: List[Thread] = []

        entries_queue = Queue()
        result_queue = Queue()

        for th_id in range(threads_count):
            th = Thread(target=_download_thread_wrapper, 
                        args=(th_id, entries_queue, result_queue, partial_progress), 
                        daemon=True, 
                        name=f"Download Thread {th_id}")
            th.start()
            threads.append(th)
        
        result_count = 0

        for entry in self.entries:
            entries_queue.put(entry)

        crash = None
        
        while result_count < entries_count:

            result = result_queue.get()
            if isinstance(result, _DownloadThreadCrash):
                crash = result
                break

            if not isinstance(result, DownloadResultProgress) or result.done:
                result_count += 1
            
            yield result_count, result

        # Send 'threads_count' sentinels.
        # We intentionally don't join thread because it takes some time for unknown 
        # reason. And we don't care of these threads because these are daemon ones.
        for th_id in range(threads_count):
            entries_queue.put(None)

        if crash is not None:
            raise ValueError(f"unexpected crash from thread {crash.thread_id}", crash.origin)


class _DownloadThreadCrash:
    """Unexpected exception happening in a thread, this is the result of a bad logic
    from programmer.
    """
    __slots__ = "thread_id", "origin",
    def __init__(self, thread_id: int, origin: Optional[Exception]) -> None:
        self.thread_id = thread_id
        self.origin = origin


def _download_thread_wrapper(
    thread_id: int, 
    entries_queue: Queue,
    result_queue: Queue,
    partial_progress: bool
) -> None:
    """Wrapper for the download thread that basically ensures that any unexpected error
    sends a signal (DownloadThreadCrash) to the master to signal the crash.
    """
    try:
        _download_thread(thread_id, entries_queue, result_queue, partial_progress)
    except Exception as e:
        result_queue.put(_DownloadThreadCrash(thread_id, e))
    except:
        # Really bad error, should not happen, but do we ever know...
        result_queue.put(_DownloadThreadCrash(thread_id, None))
        raise


def _download_thread(
    thread_id: int, 
    entries_queue: Queue,
    result_queue: Queue,
    partial_progress: bool
) -> None:
    """This function is internally used for multi-threaded download.

    :param entries_queue: Where entries to download are received.
    :param result_queue: Where threads send progress update.
    """
    
    # Cache for connections depending on host and https
    conn_cache: Dict[Tuple[bool, str], Union[HTTPConnection, HTTPSConnection]] = {}

    # Each thread has its own buffer.
    buffer_cap = 65536
    buffer_back = bytearray(buffer_cap)
    buffer = memoryview(buffer_back)

    # Try to use certifi if installed, we use this context for the connections.
    try:
        import certifi
        ctx = ssl.create_default_context(cafile=certifi.where())
    except ImportError:
        ctx = None
    
    # Maximum tries count or a single entry.
    max_try_count = 3
    max_redirect = 10

    # For speed calculation.
    speed_update_interval = 0.25
    speed_smoothing = 0.3
    speed_last_time = 0
    speed_last_size = 0
    speed_current_size = 0
    speed = 0

    while True:

        raw_entry: Optional[_DownloadEntry] = entries_queue.get()

        # None is a sentinel to stop the thread, it should be consumed ONCE.
        if raw_entry is None:
            break

        conn_key = (raw_entry.https, raw_entry.host)
        entry = raw_entry.entry

        # Get connection from cache or create it.
        conn = conn_cache.get(conn_key)
        
        last_error: Optional[str] = None
        last_error_origin: Optional[Exception] = None
        try_num = 0

        while True:

            try_num += 1
            if try_num > max_try_count:
                # Retrying implies that we have set an error.
                assert last_error is not None
                result_queue.put(DownloadResultError(thread_id, entry, last_error, last_error_origin))
                break

            # If there is no cached connection or the connection has been reset.
            if conn is None:
                if raw_entry.https:
                    conn = HTTPSConnection(raw_entry.host, raw_entry.port, context=ctx)
                else:
                    conn = HTTPConnection(raw_entry.host, raw_entry.port)
                # Cache the connection for later use.
                conn_cache[conn_key] = conn
            
            # This try-except block is around all potential 
            try:
                
                conn.request("GET", raw_entry.path)
                res = conn.getresponse()

                if res.status != 200:

                    # This loop is used to skip all bytes in the stream, 
                    # and allow further request.
                    while res.readinto(buffer):
                        pass
                    
                    if res.status == 301 or res.status == 302:
                        
                        if raw_entry.redirect >= max_redirect:
                            last_error = DownloadResultError.TOO_MANY_REDIRECT
                            continue

                        # If location header is absent, consider it not found.
                        redirect_url = res.headers.get("location")
                        if redirect_url is not None:
                            redirect_entry = DownloadEntry(
                                redirect_url, 
                                entry.dst, 
                                size=entry.size, 
                                sha1=entry.sha1, 
                                name=entry.name)
                            
                            entries_queue.put(_DownloadEntry.from_entry(redirect_entry, redirect=raw_entry.redirect + 1))
                            break  # Abort on redirect

                    # Any other non-200 code is considered not found and we retry...
                    last_error = DownloadResultError.NOT_FOUND
                    continue
                
                sha1 = None if entry.sha1 is None else hashlib.sha1()
                size = 0

                entry.dst.parent.mkdir(parents=True, exist_ok=True)
                with entry.dst.open("wb") as dst_fp:

                    while True:

                        read_len = res.readinto(buffer)
                        if not read_len:
                            break

                        size += read_len
                        speed_current_size += read_len
                        buffer_view = buffer[:read_len]
                        if sha1 is not None:
                            sha1.update(buffer_view)
                        dst_fp.write(buffer_view)

                        # Update speed calculation at given interval.
                        now = time.monotonic()
                        speed_elapsed_time = now - speed_last_time
                        if speed_elapsed_time > speed_update_interval:
                            speed_elapsed_size = speed_current_size - speed_last_size
                            current_speed = speed_elapsed_size / speed_elapsed_time
                            speed = speed_smoothing * current_speed + (1 - speed_smoothing) * speed
                            speed_last_time = now
                            speed_last_size = speed_current_size

                        # Filled the whole buffer, send a progress update because we'll 
                        # likely need another reading.
                        if partial_progress and read_len == buffer_cap:
                            result_queue.put(DownloadResultProgress(
                                thread_id,
                                entry,
                                size,
                                speed,
                                False
                            ))
            
                # If the entry should be executable, only those that can read would be
                # able to execute it.
                if entry.executable:
                    prev_mode = entry.dst.stat().st_mode
                    entry.dst.chmod(prev_mode | ((prev_mode & 0o444) >> 2))

                # Checking size and sha1 if relevant, if no error we send the full 
                # progress.
                if entry.size is not None and size != entry.size:
                    last_error = DownloadResultError.INVALID_SIZE
                elif sha1 is not None and sha1.hexdigest() != entry.sha1:
                    last_error = DownloadResultError.INVALID_SHA1
                else:
                    
                    result_queue.put(DownloadResultProgress(
                        thread_id,
                        entry,
                        size,
                        speed,
                        True
                    ))

                    # Breaking means success, to avoid unlinking the file!
                    break

            except (ConnectionError, OSError, HTTPException) as e:

                # On errors, we just throw away the old connection and create a new one.
                # Raw but efficient way of resetting the potentially broken state...
                conn.close()
                conn = None

                last_error = DownloadResultError.CONNECTION
                last_error_origin = e

            # We are here only when the file download has started but checks have failed,
            # then we should remove the file.
            try:
                entry.dst.unlink()
            except FileNotFoundError:
                pass  # Not a problem if the file isn't present.
